from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Optional
from weakref import finalize, ref

from opencosmo.index.protocols import DataIndex
from opencosmo.index.take import take

if TYPE_CHECKING:
    import numpy as np

    from opencosmo.index import DataIndex


def finish(
    cached_data: dict[str, np.ndarray],
    index: DataIndex,
    cache_ref: ref[ColumnCache],
):
    cache = cache_ref()
    if cache is None:
        return

    columns_to_add = (
        cache.registered_columns.intersection(cached_data.keys()) - cache.columns
    )
    data = {name: index.get_data(cached_data[name]) for name in columns_to_add}
    if data:
        cache.add_data(data)


def check_length(cache: ColumnCache, data: dict[str, np.ndarray]):
    lengths = set(len(d) for d in data.values())
    if len(lengths) > 1:
        raise ValueError(
            "When adding data to the cache, all columns must be the same length"
        )
    elif (l := len(cache)) > 0 and l != lengths.pop():
        raise ValueError(
            "When adding data to the cache, the columns must be the same length as the columns currently in the cache"
        )


class ColumnCache:
    """
    A column cache is used to persist data that is read from an hdf5 file. Caches can get data in one of two ways:
    1. They are explicitly given data that has been recently read from disk or
    2. They take data from a previous cache

    ColumnCaches break some of the rules that most other things follow in this library, notably that they have internal
    state (which can change). This mutability is required for two reasons.

    1. If the parent cache is garbage collected, the child cache needs to be able to copy over any data it needs
    2. If a new cache is created by adding columns, we need to signal the child to update their parent to the new
       cache. This allows us to preserve the standard "operations create new objects" pattern that is present
       throughout the library.

    """

    def __init__(
        self,
        cached_data: dict[str, np.ndarray],
        registered_column_groups: dict[int, set[str]],
        derived_index: Optional[DataIndex] = None,
        parent: Optional[ref[ColumnCache]] = None,
        children: Optional[list[ref[ColumnCache]]] = None,
    ):
        self.__cached_data = cached_data
        self.__registered_column_groups = registered_column_groups
        self.__derived_index = derived_index
        self.__parent = parent
        if children is None:
            children = []
        self.__children = children
        self.__finalizer = None

        if parent is not None and (p := parent()) is not None:
            assert derived_index is not None
            self.__finalizer = finalize(
                p,
                finish,
                p.__cached_data,
                derived_index,
                ref(self),
            )
            self.__finalizer.atexit = False  # type: ignore

    @classmethod
    def empty(cls):
        return ColumnCache({}, {}, None, None, [])

    @property
    def columns(self):
        return set(self.__cached_data.keys())

    @property
    def registered_columns(self):
        return set().union(*list(self.__registered_column_groups.values()))

    def __push_down(self, data: dict[str, np.ndarray]):
        all_columns = set().union(*list(self.__registered_column_groups.values()))
        columns_to_keep = all_columns.intersection(data.keys()).difference(
            self.__cached_data.keys()
        )
        assert self.__derived_index is not None
        for column in columns_to_keep:
            self.__cached_data[column] = self.__derived_index.get_data(data[column])

    def register_column_group(self, key: int, data: set[str]):
        assert key not in self.__registered_column_groups
        self.__registered_column_groups[key] = data

    def deregister_column_group(self, state_id: int):
        assert state_id in self.__registered_column_groups
        columns = self.__registered_column_groups.pop(state_id)
        remaining_columns = set().union(*list(self.__registered_column_groups.values()))

        to_drop = columns.difference(remaining_columns)
        cached_data = {
            name: self.__cached_data.pop(name)
            for name in to_drop
            if name in self.__cached_data
        }
        if not cached_data:
            return

        for child_ in self.__children:
            if (child := child_()) is None:
                continue
            child.__push_down(cached_data)

    def __update_parent(self, parent: ColumnCache):
        assert self.__parent is not None
        assert self.__derived_index is not None
        assert self.__finalizer is not None
        self.__finalizer.detach()
        self.__parent = ref(parent)
        if len(parent) == 8 and self.__derived_index.range()[1] >= 8:
            raise ValueError
        self.__finalizer = finalize(
            parent, finish, parent.__cached_data, self.__derived_index, ref(self)
        )
        self.__finalizer.atexit = False  # type: ignore

    def __len__(self):
        if not self.__cached_data and self.__derived_index is None:
            return 0
        elif self.__derived_index is not None:
            return len(self.__derived_index)
        else:
            return len(next(iter(self.__cached_data.values())))

    def add_data(self, data: dict[str, np.ndarray]):
        """
        The in-place equivalent of with_data. Should not be used outside the context of this
        file.

        """
        if not self.registered_columns.issuperset(data.keys()):
            raise ValueError(
                "Tried to add columns to the cache that are not registered to it!"
            )
        check_length(self, data)
        self.__cached_data = self.__cached_data | data

    def with_data(self, data: dict[str, np.ndarray]):
        if not self.registered_columns.issuperset(data.keys()):
            raise ValueError(
                "Tried to add columns to the cache that are not registered to it!"
            )
        check_length(self, data)
        new_cached_data = self.__cached_data | data
        new_cache = ColumnCache(
            new_cached_data,
            self.__registered_column_groups,
            self.__derived_index,
            self.__parent,
            self.__children,
        )
        for child_ref in self.__children:
            if (child := child_ref()) is not None:
                child.__update_parent(new_cache)
        return new_cache

    def drop(self, columns: Iterable[str]):
        columns_in_cache = set(self.__cached_data.keys()).intersection(columns)
        for column in columns_in_cache:
            del self.__cached_data[column]

    def request(self, column_names: Iterable[str], index: DataIndex):
        column_names = set(column_names)
        columns_in_cache = column_names.intersection(self.__cached_data.keys())
        missing_columns = column_names - columns_in_cache

        data = {
            name: index.get_data(self.__cached_data[name]) for name in columns_in_cache
        }
        if self.__parent is None or column_names == columns_in_cache:
            return data

        parent = self.__parent()
        if parent is None:
            return data
        assert self.__derived_index is not None
        new_index = take(self.__derived_index, index)
        return data | parent.request(column_names, new_index)

    def take(self, index: DataIndex):
        if len(self) == 0:
            return ColumnCache.empty()
        if index.range()[1] > len(self):
            raise ValueError(
                "Tried to take more elements than the length of the cache!"
            )
        new_cache = ColumnCache({}, {}, index, ref(self))
        self.__children.append(ref(new_cache))
        return new_cache

    def get_columns(self, columns: Iterable[str]):
        columns = set(columns)
        if not columns.issubset(self.registered_columns):
            raise ValueError(
                "Tried to get columns that are not registered to this cache!"
            )

        columns_in_cache = columns.intersection(self.__cached_data.keys())
        missing_columns = columns - columns_in_cache
        output = {c: self.__cached_data[c] for c in columns_in_cache}
        output |= self.__get_derived_columns(missing_columns)
        return output

    def __get_derived_columns(self, column_names: set[str]):
        if self.__derived_index is None:
            return {}
        assert self.__parent is not None
        parent = self.__parent()
        if parent is None:
            return {}
        result = parent.request(column_names, self.__derived_index)
        self.__cached_data = self.__cached_data | result
        return result



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import click

from opencosmo.analysis.install import get_file_versions, install_spec

if TYPE_CHECKING:
    from pathlib import Path


@click.group()
def cli():
    pass


@cli.command(name="install")
@click.argument("spec_name", required=True)
@click.option("--file", type=click.Path(exists=True), required=False)
@click.option("--dev", is_flag=True)
def install(spec_name: str, file: Optional[Path] = None, dev: bool = False):
    if file is not None:
        versions = get_file_versions(spec_name, file)
    else:
        versions = {}

    install_spec(spec_name, versions, dev=dev)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import operator as op
from functools import cache, partialmethod
from typing import Any, Callable, Iterable, Optional, Union

import astropy.units as u  # type: ignore
import numpy as np
from astropy import table  # type: ignore

Comparison = Callable[[float, float], bool]


class UnitsError(Exception):
    pass


def col(column_name: str) -> Column:
    """
    Create a reference to a column with a given name. These references can be combined
    to produce new columns or express queries that operate on the values in a given
    dataset. For example:

    .. code-block:: python

        import opencosmo as oc
        ds = oc.open("haloproperties.hdf5")
        query = oc.col("fof_halo_mass") > 1e14
        px = oc.col("fof_halo_mass") * oc.col("fof_halo_com_vx")
        ds = ds.with_new_columns(fof_halo_com_px = px).filter(query)

    For more advanced usage, see :doc:`cols`

    """
    return Column(column_name)


ColumnOrScalar = Union["Column", "DerivedColumn", int, float]


class Column:
    """
    A column representa a column in the table. This is used first and foremost
    for masking purposes. For example, if a user has loaded a dataset they
    can mask it with

    dataset.mask(oc.Col("column_name") < 5)

    In practice, this is just a factory class that returns masks and
    derived columns
    """

    def __init__(self, column_name: str):
        self.column_name = column_name

    # mypy doesn't reason about eq and neq correctly
    def __eq__(self, other: float | u.Quantity) -> ColumnMask:  # type: ignore
        return ColumnMask(self.column_name, other, op.eq)

    def __ne__(self, other: float | u.Quantity) -> ColumnMask:  # type: ignore
        return ColumnMask(self.column_name, other, op.ne)

    def __gt__(self, other: float | u.Quantity) -> ColumnMask:
        return ColumnMask(self.column_name, other, op.gt)

    def __ge__(self, other: float | u.Quantity) -> ColumnMask:
        return ColumnMask(self.column_name, other, op.ge)

    def __lt__(self, other: float | u.Quantity) -> ColumnMask:
        return ColumnMask(self.column_name, other, op.lt)

    def __le__(self, other: float | u.Quantity) -> ColumnMask:
        return ColumnMask(self.column_name, other, op.le)

    def isin(self, other: Iterable[float | u.Quantity]) -> ColumnMask:
        return ColumnMask(self.column_name, other, np.isin)

    def __rmul__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float():
                return self * other
            case _:
                return NotImplemented

    def __mul__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float() | Column():
                return DerivedColumn(self, other, op.mul)
            case _:
                return NotImplemented

    def __rtruediv__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float():
                return DerivedColumn(other, self, op.truediv)
            case _:
                return NotImplemented

    def __truediv__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float() | Column():
                return DerivedColumn(self, other, op.truediv)
            case _:
                return NotImplemented

    def __pow__(self, other: Any) -> DerivedColumn:
        match other:
            case int() | float():
                return DerivedColumn(self, other, op.pow)
            case _:
                return NotImplemented

    def __add__(self, other: Any) -> DerivedColumn:
        match other:
            case Column():
                return DerivedColumn(self, other, op.add)
            case _:
                return NotImplemented

    def __sub__(self, other: Any) -> DerivedColumn:
        match other:
            case Column():
                return DerivedColumn(self, other, op.sub)
            case _:
                return NotImplemented


class DerivedColumn:
    """
    A derived column represents a combination of multiple columns that already exist in
    the dataset through multiplication or division by other columns or scalars, which
    may or may not have units of their own.

    In general this is dangerous, because we cannot necessarily infer how a particular
    unit is supposed to respond to unit transformations. For the moment, we only allow
    for combinations of columns that already exist in the dataset.

    In general, columns that exist in the dataset are materialized first. Derived
    columns are then computed from these. The order of creation of the derived columns
    must be kept constant, in case you get another column which is derived from a
    derived column.
    """

    def __init__(
        self,
        lhs: ColumnOrScalar,
        rhs: ColumnOrScalar,
        operation: Callable,
        description: Optional[str] = None,
    ):
        self.lhs = lhs
        self.rhs = rhs
        self.operation = operation
        self.description = description

    def check_parent_existance(self, names: set[str]):
        match self.rhs:
            case Column():
                rhs_valid = self.rhs.column_name in names
            case DerivedColumn():
                rhs_valid = self.rhs.check_parent_existance(names)
            case _:
                rhs_valid = True

        match self.lhs:
            case Column():
                lhs_valid = self.lhs.column_name in names
            case DerivedColumn():
                lhs_valid = self.lhs.check_parent_existance(names)
            case _:
                lhs_valid = True

        return lhs_valid and rhs_valid

    def get_units(self, units: dict[str, u.Unit]):
        match self.lhs:
            case Column():
                lhs_unit = units[self.lhs.column_name]
            case DerivedColumn():
                lhs_unit = self.lhs.get_units(units)
            case _:
                lhs_unit = None
        match self.rhs:
            case Column():
                rhs_unit = units[self.rhs.column_name]
            case DerivedColumn():
                rhs_unit = self.rhs.get_units(units)
            case _:
                rhs_unit = None

        if self.operation in (op.sub, op.add):
            if lhs_unit != rhs_unit:
                raise UnitsError("Cannot add/subtract columns with different units!")
            return lhs_unit

        match (lhs_unit, rhs_unit):
            case (None, None):
                return None
            case (_, None):
                if self.operation == op.pow:
                    return self.operation(lhs_unit, self.rhs)
                else:
                    return self.operation(lhs_unit, 1)
            case (None, _):
                return self.operation(1, rhs_unit)
            case (_, _):
                return self.operation(lhs_unit, rhs_unit)

    @cache
    def requires(self):
        """
        Return the raw data columns required to make this column
        """
        vals = set()
        match self.lhs:
            case Column():
                vals.add(self.lhs.column_name)
            case DerivedColumn():
                vals = vals | self.lhs.requires()
        match self.rhs:
            case Column():
                vals.add(self.rhs.column_name)
            case DerivedColumn():
                vals = vals | self.rhs.requires()

        return vals

    def combine_on_left(self, other: Column | DerivedColumn, operation: Callable):
        """
        Combine such that this column becomes the lhs of a new derived column.
        """
        match other:
            case Column() | DerivedColumn() | int() | float():
                return DerivedColumn(self, other, operation)
            case _:
                return NotImplemented

    def combine_on_right(self, other: Column | DerivedColumn, operation: Callable):
        """
        Combine such that this column becomes the rhs of a new derived column.
        """
        match other:
            case Column() | DerivedColumn() | int() | float():
                return DerivedColumn(other, self, operation)
            case _:
                return NotImplemented

    __mul__ = partialmethod(combine_on_left, operation=op.mul)
    __rmul__ = partialmethod(combine_on_right, operation=op.mul)
    __truediv__ = partialmethod(combine_on_left, operation=op.truediv)
    __rtruediv__ = partialmethod(combine_on_right, operation=op.truediv)
    __pow__ = partialmethod(combine_on_left, operation=op.pow)
    __add__ = partialmethod(combine_on_left, operation=op.add)
    __radd__ = partialmethod(combine_on_right, operation=op.add)
    __sub__ = partialmethod(combine_on_left, operation=op.sub)
    __rsub__ = partialmethod(combine_on_right, operation=op.sub)

    def evaluate(self, data: table.Table) -> table.Column:
        match self.lhs:
            case DerivedColumn():
                lhs = self.lhs.evaluate(data)
            case Column():
                lhs = data[self.lhs.column_name]
            case _:
                lhs = self.lhs
        match self.rhs:
            case DerivedColumn():
                rhs = self.rhs.evaluate(data)
            case Column():
                rhs = data[self.rhs.column_name]
            case _:
                rhs = self.rhs

        result = self.operation(lhs, rhs)
        return result


class ColumnMask:
    """
    A mask is a class that represents a mask on a column. ColumnMasks evaluate
    to t/f for every element in the given column.
    """

    def __init__(
        self,
        column_name: str,
        value: float | u.Quantity,
        operator: Callable[[table.Column, float | u.Quantity], np.ndarray],
    ):
        self.column_name = column_name
        self.value = value
        self.operator = operator

    def apply(self, column: u.Quantity | np.ndarray) -> np.ndarray:
        """
        mask the dataset based on the mask.
        """
        # Astropy's errors are good enough here
        if isinstance(column, table.Table):
            column = column[self.column_name]

        if isinstance(self.value, u.Quantity) and isinstance(column, u.Quantity):
            if self.value.unit != column.unit:
                raise ValueError(
                    f"Incompatible units in fiter: {self.value.unit} and {column.unit}"
                )

        elif isinstance(column, u.Quantity):
            return self.operator(column.value, self.value)

        return self.operator(column, self.value)  # type: ignore



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from warnings import warn

import pyarrow  # type: ignore
from pyarrow import parquet as pq

import opencosmo as oc

if TYPE_CHECKING:
    from os import PathLike

    from opencosmo.collection.protocols import Collection


def write_parquet(
    path: PathLike,
    to_write: oc.Dataset | Collection,
    overwrite: bool = False,
    **kwargs,
):
    """
    Write a dataset or collection to a parquet file at the given path. If you are writing a :py:class:`opencosmo.Dataset`,
    or :py:class:`opencosmo.Lightcone` the data will be written as a single file at the given path. If you are writing
    a :py:class:`opencosmo.StructureCollection` the data will be written to several files, one for each type of particle.

    Parameters
    ----------
    path: PathLike
        The path to write the data to. If you are writing a :py:class:`Dataset  <opencosmo.Dataset>`
        or :py:class:`Lightcone <opencosmo.Lightcone>` this should be a single parquet file. If you
        are writing a :py:class:`StructureCollection <opencosmo.StructureCollection>`, this should be
        a folder.

    to_write: opencosmo.Dataset | opencosmo.Lightcone | opencosmo.StructureCollection
        The dataset or collection to write

    overwrite: bool, default = False
        If true, any existing data at the given path will be overwritten.


    """
    if not isinstance(path, Path):
        path = Path(path)
    match to_write:
        case oc.Dataset() | oc.Lightcone():
            return __write_dataset(path, to_write, overwrite, **kwargs)
        case oc.StructureCollection():
            return __write_structure_collection(path, to_write, overwrite, **kwargs)
        case _:
            raise ValueError(f"No parquet writer defined for type {type(to_write)}")


def __write_dataset(path: Path, dataset: oc.Dataset | oc.Lightcone, overwrite: bool):
    data = dataset.get_data("numpy")
    if path.exists() and not overwrite:
        raise FileExistsError(path)
    output = {}
    if not isinstance(data, dict):
        raise NotImplementedError
    for name, column in data.items():
        if column.dtype.names is None:
            output[name] = column
            continue
        outputs = {f"{name}_{cname}": column[cname] for cname in column.dtype.names}
        output.update(outputs)

    table = pyarrow.table(output)
    pq.write_table(table, path)


def __write_structure_collection(
    path: Path, collection: oc.StructureCollection, overwrite: bool
):
    if path.exists() and not path.is_dir():
        raise ValueError(
            "Dumping a structure collection to parquet can create multiple files, so the output location should be a directory"
        )

    elif not path.exists():
        path.mkdir(parents=True)

    for name, dataset in collection.items():
        if "particle" not in name:
            warn(f"write_parquet only supports writing particles, skipping {name}")
            continue
        assert isinstance(dataset, oc.Dataset)
        __write_dataset(path / f"{name}.parquet", dataset, overwrite)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union

import numpy as np
import yt  # type: ignore
from pyxsim import CIESourceModel  # type: ignore
from unyt import unyt_array, unyt_quantity  # type: ignore

if TYPE_CHECKING:
    from yt.data_objects.static_output import Dataset as YT_Dataset

    import opencosmo as oc

# ---- define some constants ---- #
mp = unyt_quantity(1.67262192595e-24, "g")
kB = unyt_quantity(1.380649e-16, "erg/K")
solar_metallicity = 0.012899  # value used internally in HACC
# ------------------------------- #


def create_yt_dataset(
    data: Dict[str, oc.Dataset],
    compute_xray_fields: Optional[bool] = False,
    return_source_model: Optional[bool] = False,
    source_model_kwargs: Optional[Dict[str, Any]] = {},
) -> Union[YT_Dataset, Tuple[YT_Dataset, CIESourceModel]]:
    """
    Converts particle data to a `yt` dataset. Note that `yt`
    is generally developed with AMR codes in mind, but support for
    SPH codes is continually being added. `yt's` documentation can
    be found `here <https://yt-project.org/doc/index.html>`_.

    If `compute_xray_fields` is enabled, X-ray emissivity and luminosity fields
    will be added using `pyxsim <https://hea-www.cfa.harvard.edu/~jzuhone/pyxsim/index.html#>`_,
    which generates photon samples from gas properties.

    Parameters
    ----------
    data : dict of astropy.table.Table
        A dictionary of particle datasets. Must include at least positions and masses.
    compute_xray_fields : bool, optional
        Whether or not to compute X-ray luminosities with `pyxsim`.
        Uses `CIESourceModel`, which considers thermal emission from gas assuming
        collisional ionization equilibrium.
    return_source_model : bool, optional
        Whether or not to return the `pyxsim` source model for further interaction,
        such as computing additional luminosities in different frequency bands
        or generating synthetic observations.
    source_model_kwargs : dict, optional
        Keyword arguments passed to the `CIESourceModel` constructor in `pyxsim`.
        These can include parameters like `emin`, `emax`, `nbins`, `abund_table`, etc.,
        to control the spectral resolution and emission model behavior. If `None`,
        default values will be used for all source model parameters.

    Returns
    -------
    ds : yt.data_objects.static_output.Dataset
        A `yt` dataset built from the input particle data, with additional fields
        (e.g., X-ray luminosities) if requested.

    source_model : pyxsim.source_models.CIESourceModel, optional
        Returned only if `return_source_model=True`.
    """

    data_dict: Dict[
        Union[Tuple[str, str], str], Union[np.ndarray, Tuple[np.ndarray, str]]
    ] = {}

    minx, maxx = np.inf, -np.inf
    miny, maxy = np.inf, -np.inf
    minz, maxz = np.inf, -np.inf

    # Fields that we need to rename to hook up to yt's internals
    special_fields = {
        "x": "particle_position_x",
        "y": "particle_position_y",
        "z": "particle_position_z",
        "mass": "particle_mass",
        "rho": "density",
        "hh": "smoothing_length",
    }

    def astropy_to_yt(array):
        """
        Converts from astropy format to yt format.
        Basically just reformats the units.
        """

        if array.unit is None:
            return unyt_array(array.data, "dimensionless")

        return unyt_array.from_astropy(array)

    for ptype in data.keys():
        if "particles" not in ptype:
            continue

        particle_data = data[ptype].data
        redshift = data[ptype].redshift
        cosmo = data[ptype].cosmology
        ptype_short = ptype.split("_")[0]

        for field in particle_data.keys():
            yt_field_name = special_fields.get(field, field)
            yt_particle_data = astropy_to_yt(particle_data[field])

            data_dict[(ptype_short, yt_field_name)] = (
                np.asarray(yt_particle_data.d),
                str(yt_particle_data.units),
            )

        minx, maxx = (
            min(minx, min(particle_data["x"].value)),
            max(maxx, max(particle_data["x"].value)),
        )
        miny, maxy = (
            min(miny, min(particle_data["y"].value)),
            max(maxy, max(particle_data["y"].value)),
        )
        minz, maxz = (
            min(minz, min(particle_data["z"].value)),
            max(maxz, max(particle_data["z"].value)),
        )

    bbox = [[minx, maxx], [miny, maxy], [minz, maxz]]

    # Raise error if any bound is still infinite
    if any(np.isinf(bound) for axis in bbox for bound in axis):
        raise ValueError(
            "Bounding box coordinates contain infinite values."
            "Check input data for missing or invalid positions."
        )

    ds = yt.load_particles(
        data_dict,
        length_unit="Mpc",
        mass_unit="Msun",
        bbox=bbox,
        periodicity=(False, False, False),
    )

    ds.sph_smoothing_style = "gather"  # seems to give more reliable results

    # set cosmology parameters

    ds.cosmological_simulation = 1
    ds.current_redshift = redshift
    ds.hubble_constant = 0.01 * cosmo.H0.value
    ds.omega_matter = cosmo.Om0
    ds.omega_lambda = cosmo.Ode0
    ds.omega_curvature = cosmo.Ok0
    ds.omega_radiation = cosmo.Onu0 + cosmo.Ogamma0

    if ("gas", "density") in ds.field_list:
        # if hydro sim, add derived fields

        # compute a new MMW field
        ds.add_field(
            ("gas", "MMW"),
            function=_mmw,
            units="",
            sampling_type="particle",
        )

        ds.add_field(
            ("gas", "temperature"),
            function=_temperature,
            units="K",
            sampling_type="particle",
        )

        ds.add_field(
            ("gas", "number_density"),
            function=_number_density,
            units="cm**-3",
            sampling_type="particle",
            force_override=True,
        )

        ds.add_field(
            ("gas", "xh"), function=_h_fraction, units="", sampling_type="particle"
        )

        ds.add_field(
            ("gas", "metallicity"),
            function=_metallicity,
            units="Zsun",
            sampling_type="particle",
        )

        if compute_xray_fields:
            # compute xray luminosities, emissivities, etc. using pyxsim.
            # This calls CIESourceModel, which assumes ionization equilibrium.
            # User can define custom parameters

            ds.add_field(
                ("gas", "emission_measure"),
                function=_emission_measure,
                units="cm**-3",
                sampling_type="particle",
            )

            default_kwargs = {
                "model": "apec",
                "emin": 0.1,  # keV
                "emax": 10.0,  # keV
                "nbins": 1000,
                "Zmet": ("gas", "metallicity"),  # Zsun
                "temperature_field": ("gas", "temperature"),
                "emission_measure_field": ("gas", "emission_measure"),
                "h_fraction": "xh",
            }

            if source_model_kwargs is None:
                source_model_kwargs = {}

            # update with user-defined settings
            source_model_kwargs = {**default_kwargs, **source_model_kwargs}

            # define xray source model (
            # NOTE: this will download a few fits files needed for the analysis)
            source = CIESourceModel(**source_model_kwargs)

            # populate yt dataset with xray fields
            source.make_source_fields(
                ds, source_model_kwargs["emin"], source_model_kwargs["emax"]
            )

            if return_source_model:
                return ds, source

    elif compute_xray_fields:
        raise RuntimeError(
            "`compute_xray_fields` can only be used with hydrodynamic simulations"
        )

    return ds


# ---------- DERIVED FIELDS -------------- #


def _mmw(field, data):
    # Recompute mean molecular weight. The "mu" field currently stored
    # is 1.0 with units of kg, which is wrong.
    Y = data["gas", "yhe"].d
    X = 1 - Y
    Z = data["gas", "zmet"].d * solar_metallicity

    # MMW for fully ionized gas
    return 1 / (2 * X + 0.75 * Y + Z / (2 * 16))


def _temperature(field, data):
    gamma = 5 / 3
    return (
        data["gas", "MMW"] * mp * data["gas", "uu"].to("cm**2/s**2") / kB * (gamma - 1)
    )


def _number_density(field, data):
    return data["gas", "density"].to("g/cm**3") / (data["gas", "MMW"] * mp)


def _metallicity(field, data):
    # metallicity in solar units
    return data["gas", "zmet"]


def _h_fraction(field, data):
    return 1 - data["gas", "yhe"]


def _emission_measure(field, data):
    # emission_measure = ne**2 * particle_volume

    # assume gas is fully ionized -- safe assumption for cluster scale objects
    ne = (1 - 0.5 * data["gas", "yhe"].d) * data["gas", "density"].to("g/cm**3") / mp
    nH = (1 - data["gas", "yhe"].d) * data["gas", "density"].to("g/cm**3") / mp

    return ne * nH * (data["gas", "particle_mass"] / data["gas", "density"]).to("cm**3")



from __future__ import annotations

from typing import TYPE_CHECKING

from opencosmo.parameters import hacc

if TYPE_CHECKING:
    from pydantic import BaseModel

    from opencosmo.parameters.file import FileParameters


def get_dtype_parameters(file_parameters: FileParameters) -> dict[str, type[BaseModel]]:
    if file_parameters.origin == "HACC":
        known_dtype_params = hacc.DATATYPE_PARAMETERS
    else:
        raise ValueError(f"Unknown dataset origin {file_parameters.origin}")
    dtype_parameters = known_dtype_params[str(file_parameters.data_type)]
    if file_parameters.is_lightcone:
        lightcone_parameters = hacc.LightconeParams
        dtype_parameters.update({"lightcone": lightcone_parameters})
    return dtype_parameters



from __future__ import annotations

from typing import TYPE_CHECKING

import numba as nb  # type: ignore
import numpy as np
from numpy.typing import ArrayLike

from . import ChunkedIndex, SimpleIndex

if TYPE_CHECKING:
    from opencosmo.index.protocols import DataIndex


def take(from_: DataIndex, by: DataIndex):
    match (from_, by):
        case (SimpleIndex(), SimpleIndex()):
            return take_simple_from_simple(from_, by)
        case (SimpleIndex(), ChunkedIndex()):
            return take_chunked_from_simple(from_, by)
        case (ChunkedIndex(), SimpleIndex()):
            return take_simple_from_chunked(from_, by)
        case (ChunkedIndex(), ChunkedIndex()):
            return take_chunked_from_chunked(from_, by)


def take_simple_from_chunked(from_: ChunkedIndex, by: SimpleIndex):
    cumulative = np.insert(np.cumsum(from_.sizes), 0, 0)[:-1]
    arr = by.into_array()

    indices_into_chunks = np.argmax(arr[:, np.newaxis] < cumulative, axis=1) - 1
    output = arr - cumulative[indices_into_chunks] + from_.starts[indices_into_chunks]
    return SimpleIndex(output)


def take_simple_from_simple(from_: SimpleIndex, by: SimpleIndex):
    return SimpleIndex(from_.into_array()[by.into_array()])


def take_chunked_from_simple(from_: SimpleIndex, by: ChunkedIndex):
    from_arr = from_.into_array()
    starts = by.starts
    sizes = by.sizes
    output = np.zeros(sizes.sum(), dtype=int)
    output = __cfs_helper(from_arr, starts, sizes, output)
    return SimpleIndex(output)


@nb.njit
def __cfs_helper(arr, starts, sizes, storage):
    rs = 0
    for i in range(len(starts)):
        cstart = starts[i]
        csize = sizes[i]
        storage[rs : rs + csize] = arr[cstart : cstart + csize]
        rs += csize
    return storage


def take_chunked_from_chunked(from_: ChunkedIndex, by: ChunkedIndex):
    if from_.is_single_chunk() and from_.range()[0] == 0:
        return by

    from_idx = from_.into_array()
    output = np.zeros(len(by), dtype=int)
    output = __cfs_helper(from_idx, by.starts, by.sizes, output)
    diff = np.diff(output)
    breaks = np.where(diff != 1)[0]

    # Start indices of segments
    starts = output[np.r_[0, breaks + 1]]
    # End indices of segments (inclusive)
    ends = output[np.r_[breaks, output.size - 1]]

    sizes = ends - starts + 1
    return ChunkedIndex(starts, sizes)



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol, Self

if TYPE_CHECKING:
    import h5py
    import numpy as np
    from numpy.typing import NDArray


class DataIndex(Protocol):
    """
    A DataIndex is responsible for tracking which elements of some hdf5 dataset
    should be read. Note these are NOT a spatial index, though a spatial index
    does return a DataIndex when it is queried.
    """

    @classmethod
    def from_size(cls, size: int) -> Self: ...
    def get_data(self, data: h5py.Dataset | np.ndarray) -> np.ndarray: ...
    def n_in_range(
        self, start: NDArray[np.int_], end: NDArray[np.int_]
    ) -> NDArray[np.int_]: ...
    def intersection(self, other: DataIndex) -> DataIndex: ...
    def into_mask(self) -> NDArray[np.bool_]: ...
    def into_array(self) -> NDArray[np.int_]: ...
    def mask(self, mask: np.ndarray) -> DataIndex: ...
    def range(self) -> tuple[int, int]: ...
    def concatenate(self, *others: DataIndex) -> DataIndex: ...
    def __len__(self) -> int: ...
    def __getitem__(self, item: int) -> DataIndex: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Protocol

if TYPE_CHECKING:
    import h5py

try:
    from mpi4py import MPI
except ImportError:
    MPI = None  # type: ignore


class DataSchema(Protocol):
    """
    A DataSchema describes the layout of a file. Like HDF5, the schemas are organized
    hierarchicaly. As a result most schemas are only responsible for holding children
    and performing any verification that must be done across multiple children. For
    example, a DatasetSchema holds ColumnSchemas, and is responsible for verifying that
    all of its columns have the same length.

    A schema most be capable of transforming into a writer. As such, some schemas such
    as the ColumnSchema must hold reference to the data that wthey will write, as well
    as an index into the data of the elements that should be included.

    Schemas also must allocate their respective structures, i.e. by creating groups or
    datasets.
    """

    def add_child(self, child: "DataSchema", child_id: Any): ...
    def allocate(self, group: h5py.File | h5py.Group): ...
    def verify(self): ...
    def into_writer(self, comm: Optional["MPI.Comm"]): ...


class DataWriter(Protocol):
    """
    Because DataSchemas are responsible for allocating files and producing the
    structure, a writer can always assume that the appropriate group or dataset already
    exists. If it doesn't exist, that is an error elsewhere that shoulds be resolved.
    """

    def write(self, group: h5py.File | h5py.Group): ...


class Writeable(Protocol):
    """
    In order to be writeable, an object must define a single method.
    """

    def make_schema(self) -> DataSchema: ...



from pydantic import BaseModel, ConfigDict, field_serializer


class BoxRegionModel(BaseModel):
    model_config = ConfigDict(frozen=True)
    p1: tuple[float, float, float]
    p2: tuple[float, float, float]


class ConeRegionModel(BaseModel):
    model_config = ConfigDict(frozen=True)
    center: tuple[float, float]
    radius: float


class HealPixRegionModel(BaseModel):
    model_config = ConfigDict(frozen=True)
    pixels: frozenset[int]
    nside: int

    @field_serializer("pixels")
    def serialize_pixels(self, value):
        return list(value)


RegionModel = BoxRegionModel | ConeRegionModel | HealPixRegionModel



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Optional

import numpy as np

if TYPE_CHECKING:
    from opencosmo.io import writers as iow


from opencosmo.mpi import MPI


def apply_updaters(
    writers: dict[str, dict[str, iow.ColumnWriter]], comm: Optional[MPI.Comm]
):
    if "data_linked" in writers:
        apply_linked_data_updaters(writers["data_linked"], comm)

    if "index" in writers:
        apply_index_updaters(writers["index"], comm)
    return writers


def apply_index_updaters(columns, comm):
    for colname, column in columns.items():
        column.updater = partial(sum_updater, comm=comm)


def sum_updater(data: np.ndarray, comm: Optional["MPI.Comm"] = None):
    if comm is not None and comm.Get_size():
        recvbuf = np.zeros_like(data)
        comm.Allreduce(data, recvbuf, MPI.SUM)
        return recvbuf
    return data


def apply_linked_data_updaters(columns, comm):
    starts = {name: col for name, col in columns.items() if "start" in name}
    sizes = {name: col for name, col in columns.items() if "size" in name}
    idxs = {name: col for name, col in columns.items() if "idx" in name}
    for name, start in starts.items():
        link_type = name.rsplit("_", maxsplit=1)[0]
        size = sizes[f"{link_type}_size"]
        start.source = size.source
        start.updater = partial(start_link_updater, comm=comm)
    for idx in idxs.values():
        idx.updater = partial(idx_link_updater, comm=comm)


def idx_link_updater(input: np.ndarray, comm) -> np.ndarray:
    output = np.full(len(input), -1)
    good = input >= 0
    if comm is not None:
        all_good = comm.allgather(sum(good))
        offsets = np.insert(np.cumsum(all_good), 0, 0)[:-1]
        offset = offsets[comm.Get_rank()]
    else:
        offset = 0

    output[good] = np.arange(sum(good)) + offset
    return output


def start_link_updater(sizes: np.ndarray, comm) -> np.ndarray:
    cumulative_sizes = np.cumsum(sizes)
    new_starts = np.insert(cumulative_sizes, 0, 0)
    if comm is not None:
        all_total_size = comm.allgather(np.sum(sizes))
        offsets = np.insert(np.cumsum(all_total_size), 0, 0)[:-1]
        offset = offsets[comm.Get_rank()]
    else:
        offset = 0

    new_starts = new_starts[:-1] + offset
    return new_starts



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Type

from astropy.units.typing import UnitLike

from opencosmo.units import UnitConvention
from opencosmo.units.get import get_unit_applicators_dict

if TYPE_CHECKING:
    from astropy.cosmology import Cosmology
    from pydantic import BaseModel

ModelUnitAnnotation = tuple[UnitConvention, dict[str, UnitLike], bool]

__KNOWN_UNITFUL_MODELS__: dict[Type[BaseModel], ModelUnitAnnotation] = {}


# Constraint: Unit covention for all fields in a given model must be the same


def register_units(
    model: Type[BaseModel],
    field_name: str,
    unit: UnitLike,
    convention: UnitConvention = UnitConvention.SCALEFREE,
    is_comoving: bool = True,
):
    model_spec = __KNOWN_UNITFUL_MODELS__.get(model)
    registered_fields: dict[str, UnitLike]
    if model_spec is not None and model_spec[0] != convention:
        raise ValueError(
            "All unitful fields in a parameter model must use the same unit convention"
        )
    elif model_spec is None:
        registered_fields = {}
    else:
        registered_fields = model_spec[1]
    if field_name in registered_fields:
        raise ValueError(f"Field {field_name} was already registered with units!")

    registered_fields[field_name] = unit
    __KNOWN_UNITFUL_MODELS__[model] = (convention, registered_fields, is_comoving)


def __get_unit_transformations(
    model: BaseModel, cosmology, convention: UnitConvention = UnitConvention.SCALEFREE
) -> dict:
    if (us := __KNOWN_UNITFUL_MODELS__.get(type(model))) is None:
        return {}
    base_convention, known_units, is_comoving = us
    applicators = get_unit_applicators_dict(
        known_units, base_convention, cosmology, is_comoving
    )
    return applicators


def apply_units(
    model: BaseModel,
    cosmology: Cosmology,
    convention: UnitConvention = UnitConvention.SCALEFREE,
    unit_kwargs: dict[str, Any] = {},
):
    applicators = __get_unit_transformations(model, cosmology, convention)
    parameters = model.model_dump()
    for name, applicator in applicators.items():
        parameters[name] = applicator.apply(
            parameters[name], convention, unit_kwargs=unit_kwargs
        )
    return parameters



from functools import cache
from typing import Optional

try:
    from mpi4py import MPI
except (ImportError, RuntimeError):
    MPI = None  # type: ignore


@cache
def get_comm_world() -> Optional["MPI.Comm"]:
    if MPI is None or MPI.COMM_WORLD.Get_size() == 1:
        return None
    return MPI.COMM_WORLD.Dup()


def get_mpi():
    return MPI



from __future__ import annotations

from collections import namedtuple
from typing import TYPE_CHECKING, NamedTuple, Type, TypeVar

if TYPE_CHECKING:
    from opencosmo import Dataset

DIFFMAH_INPUT = namedtuple(
    "DIFFMAH_INPUT", ["logm0", "logtc", "early_index", "late_index", "t_peak"]
)

T = TypeVar("T", bound=NamedTuple)


def make_named_tuple(dataset: Dataset, input_tuple: Type[T]) -> T:
    required_columns = input_tuple._fields
    data = dataset.select(required_columns).data
    output = {c: data[c].value for c in required_columns}
    return input_tuple(**output)  # type: ignore



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Callable,
    Generator,
    Iterable,
    Mapping,
    Optional,
    TypeAlias,
)
from warnings import warn

import astropy.units as u  # type: ignore
import numpy as np
from astropy.table import QTable  # type: ignore

from opencosmo.dataset.evaluate import visit_dataset
from opencosmo.index import ChunkedIndex, SimpleIndex
from opencosmo.spatial import check
from opencosmo.units.converters import get_scale_factor

if TYPE_CHECKING:
    from astropy import units
    from astropy.cosmology import Cosmology

    from opencosmo.dataset.column import ColumnMask, DerivedColumn
    from opencosmo.dataset.handler import Hdf5Handler
    from opencosmo.dataset.state import DatasetState
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.index import DataIndex
    from opencosmo.io.schemas import DatasetSchema
    from opencosmo.parameters import HaccSimulationParameters
    from opencosmo.spatial.protocols import Region
    from opencosmo.spatial.tree import Tree


OpenCosmoData: TypeAlias = QTable | u.Quantity | dict[str, np.ndarray] | np.ndarray


class Dataset:
    def __init__(
        self,
        header: OpenCosmoHeader,
        state: DatasetState,
        tree: Optional[Tree] = None,
    ):
        self.__header = header
        self.__state = state
        self.__tree = tree

    def __repr__(self):
        """
        A basic string representation of the dataset
        """
        length = len(self)

        if len(self) < 10:
            repr_ds = self
            table_head = ""
        else:
            repr_ds = self.take(10, at="start")
            table_head = "First 10 rows:\n"

        table_repr = repr_ds.data.__repr__()
        # remove the first line
        table_repr = table_repr[table_repr.find("\n") + 1 :]
        head = f"OpenCosmo Dataset (length={length})\n"
        cosmo_repr = f"Cosmology: {self.cosmology.__repr__()}" + "\n"
        return head + cosmo_repr + table_head + table_repr

    @property
    def index(self):
        return self.__state.raw_index

    def __len__(self):
        return len(self.__state)

    def __enter__(self):
        # Need to write tests
        return self

    def __exit__(self, *exc_details):
        return self.__state.__exit__(*exc_details)

    def close(self):
        return self.__state.__exit__()

    @property
    def header(self) -> OpenCosmoHeader:
        """
        The header associated with this dataset.

        OpenCosmo headers generally contain information about the original data this
        dataset was produced from, as well as any analysis that was done along
        the way.

        Returns
        -------
        header: opencosmo.header.OpenCosmoHeader

        """
        return self.__header

    @property
    def columns(self) -> list[str]:
        """
        The names of the columns in this dataset.

        Returns
        -------
        columns: list[str]
        """
        return self.__state.columns

    @property
    def meta_columns(self) -> list[str]:
        return self.__state.meta_columns

    @property
    def descriptions(self) -> dict[str, Optional[str]]:
        """
        Return the descriptions (if any) of the columns in this dataset as a dictonary.
        Columns without a description will be included in the dictionary with a value
        of None

        Returns
        -------

        descriptions : dict[str, str | None]
            The column descriptions
        """
        return self.__state.descriptions

    @property
    def cosmology(self) -> Cosmology:
        """
        The cosmology of the simulation this dataset is drawn from as
        an astropy.cosmology.Cosmology object.

        Returns
        -------
        cosmology: astropy.cosmology.Cosmology
        """
        return self.__header.cosmology

    @property
    def dtype(self) -> str:
        """
        The data type of this dataset.

        Returns
        -------
        dtype: str
        """
        return str(self.__header.file.data_type)

    @property
    def redshift(self) -> float | tuple[float, float]:
        """
        The redshift slice or range this dataset was drawn from

        Returns
        -------
        redshift: float

        """
        return self.__header.file.redshift

    @property
    def region(self) -> Region:
        """
        The region this dataset is contained in. If no spatial
        queries have been performed, this will be the entire
        simulation box for snapshots or the full sky for lightcones

        Returns
        -------
        region: opencosmo.spatial.Region

        """
        return self.__state.region

    @property
    def simulation(self) -> HaccSimulationParameters:
        """
        The parameters of the simulation this dataset is drawn
        from.

        Returns
        -------
        parameters: opencosmo.parameters.hacc.HaccSimulationParameters
        """
        return self.__header.simulation

    @property
    def data(self) -> QTable | u.Quantity:
        """
        Return the data in the dataset in astropy format. The value of this
        attribute is equivalent to the return value of
        :code:`Dataset.get_data("astropy")`.

        Returns
        -------
        data : astropy.table.Table or astropy.table.Column
            The data in the dataset.

        """
        # should rename this, dataset.data can get confusing
        # Also the point is that there's MORE data than just the table
        return self.get_data("astropy")

    def get_metadata(self, columns: list[str] = []):
        return self.__state.get_metadata(columns)

    def get_data(
        self,
        output="astropy",
        unpack=True,
        metadata_columns=[],
    ) -> OpenCosmoData:
        """
        Get the data in this dataset as an astropy table/column or as
        numpy array(s). Note that a dataset does not load data from disk into
        memory until this function is called. As a result, you should not call
        this function until you have performed any transformations you plan to
        on the data.

        You can get the data in two formats, "astropy" (the default) and "numpy".
        "astropy" format will return the data as an astropy table with associated
        units. "numpy" will return the data as a dictionary of numpy arrays. The
        numpy values will be in the associated unit convention, but no actual
        units will be attached.

        If the dataset only contains a single column, it will be returned as an
        astropy quantity (if it has units) or numpy array.

        This method does not cache data. Calling "get_data" always reads data
        from disk, even if you have already called "get_data" in the past.
        You can use :py:attr:`Dataset.data <opencosmo.Dataset.data>` to return
        data and keep it in memory.

        Parameters
        ----------
        output: str, default="astropy"
            The format to output the data in

        Returns
        -------
        data: Table | Quantity | dict[str, ndarray] | ndarray
            The data in this dataset.
        """
        if output not in {"astropy", "numpy"}:
            raise ValueError(f"Unknown output type {output}")

        if self.__state.convention.value == "physical":
            scale_factor = get_scale_factor(self.__state, self.cosmology, self.redshift)
            unit_kwargs = {"scale_factor": scale_factor}
        else:
            unit_kwargs = {}

        data = self.__state.get_data(
            unit_kwargs=unit_kwargs, metadata_columns=metadata_columns
        )  # table
        if len(data) == 1 and unpack:  # unpack length-1 tables
            data = {name: data[0] for name, data in data.items()}
        elif len(data.colnames) == 1:
            cn = data.colnames[0]
            data = data[cn]

        if output == "numpy":
            if isinstance(data, u.Quantity):
                data = data.value
            elif isinstance(data, (QTable, dict)):
                data = dict(data)
                is_quantity = filter(
                    lambda v: isinstance(data[v], u.Quantity), data.keys()
                )
                for colname in is_quantity:
                    data[colname] = data[colname].value

        if isinstance(data, dict) and len(data) == 1:
            return next(iter(data.values()))

        return data

    def bound(self, region: Region, select_by: Optional[str] = None):
        """
        Restrict the dataset to some subregion. The subregion will always be evaluated
        in the same units as the current dataset. For example, if the dataset is
        in the default "comoving" unit convention, positions are always in units of
        comoving Mpc. However Region objects themselves do not carry units.
        See :doc:`spatial_ref` for details of how to construct regions.

        Parameters
        ----------
        region: opencosmo.spatial.Region
            The region to query.

        Returns
        -------
        dataset: opencosmo.Dataset
            The portion of the dataset inside the selected region

        Raises
        ------
        ValueError
            If the query region does not overlap with the region this dataset resides
            in
        AttributeError:
            If the dataset does not contain a spatial index
        """
        if self.__tree is None:
            raise AttributeError(
                "Your dataset does not contain a spatial index, "
                "so spatial querying is not available"
            )

        if not self.header.file.is_lightcone:
            columns = check.find_coordinates_3d(self, self.dtype)

            check_region = region.into_base_convention(
                self.__state.unit_handler,
                columns,
                self.__state.convention,
                {
                    "scale_factor": self.cosmology.scale_factor(
                        self.header.file.redshift
                    ).value
                },
            )
        else:
            check_region = region

        if not self.__state.region.intersects(check_region):
            new_index = ChunkedIndex.empty()
            new_state = self.__state.take_rows(new_index)
            return Dataset(self.__header, new_state, self.__tree)

        if not self.__state.region.contains(check_region):
            warn(
                "You're querying with a region that is not fully contained by the "
                "region this dataset is in. This may result in unexpected behavior"
            )

        contained_index: DataIndex
        intersects_index: DataIndex
        contained_index, intersects_index = self.__tree.query(check_region)

        contained_index = self.__state.raw_index.projection(contained_index)
        intersects_index = self.__state.raw_index.projection(intersects_index)

        check_state = self.__state.take_rows(intersects_index)
        check_dataset = Dataset(
            self.__header,
            check_state,
            self.__tree,
        )
        if not self.__header.file.is_lightcone:
            check_dataset = check_dataset.with_units("scalefree")

        mask = check.check_containment(check_dataset, check_region, self.__header.file)
        new_intersects_index = intersects_index.mask(mask)

        new_index = contained_index.concatenate(new_intersects_index)

        new_state = self.__state.take_rows(new_index).with_region(check_region)

        return Dataset(self.__header, new_state, self.__tree)

    def evaluate(
        self,
        func: Callable,
        vectorize=False,
        insert=False,
        format="astropy",
        **evaluate_kwargs,
    ) -> Dataset | np.ndarray:
        """
        Iterate over the rows in this dataset, apply :code:`func` to each, and collect
        the result as new columns in the dataset.

        This function is the equivalent of :py:meth:`with_new_columns <opencosmo.Dataset.with_new_columns>`
        for cases where the new column is not a simple algebraic combination of existing columns. Unlike
        :code:`with_new_columns`, this method will evaluate the results immediately and the resulting
        columns will not change under unit transformations. You may also choose to simply return the result
        instead of adding it as a column.

        The function should take in arguments with the same name as the columns in this dataset that
        are needed for the computation, and should return a dictionary of output values.
        The dataset will automatically selected the needed columns to avoid reading unnecessarily reading
        data from disk. You may also include all columns in the dataset by providing a function with a single
        import argument with the same name as the data type of this dataset (see :py:attr:`Dataset.dtype <opencosmo.Dataset.dtype>`
        In this case, the data will be provided as a dictionary of astropy quantity arrays or numpy arrays

        The new columns will have the same names as the keys of the output dictionary
        See :ref:`Evaluating On Datasets` for more details.

        If vectorize is set to True, the full columns will be pased to the dataset. Otherwise,
        rows will be passed to the function one at a time.

        If the function returns None, this method will also return None as output. For example, the function
        could simply produce plots and save the to files.

        Parameters
        ----------

        func: Callable
            The function to evaluate on the rows in the dataset.

        vectorize: bool, default = False
            Whether to provide the values as full columns (True) or one row at a time (False)

        insert: bool, default = True
            If true, the data will be inserted as a column in this dataset. The new column will have the same name
            as the function. Otherwise the data will be returned directly.

        format: str, default = astropy
            Whether to provide data to your function as "astropy" quantities or "numpy" arrays/scalars. Default "astropy"

        **evaluate_kwargs: any,
            Any additional arguments that are required for your function to run. These will be passed directly
            to the function as keyword arguments. If a kwarg is an array of values with the same length as the dataset,
            it will be treated as an additional column.

        Returns
        -------
        result : Dataset | dict[str, np.ndarray | astropy.units.Quantity]
            The new dataset with the evaluated column(s) or the results as numpy arrays or astropy quantities
        """
        kwarg_columns = set(evaluate_kwargs.keys()).intersection(self.columns)
        if kwarg_columns:
            raise ValueError(
                "Keyword arguments cannot have the same name as columns in your dataset!"
            )

        output = visit_dataset(func, self, vectorize, format, evaluate_kwargs)
        if output is None or not insert:
            return output
        is_same_length = all(
            isinstance(o, np.ndarray) and len(o) == len(self) for o in output.values()
        )

        if not is_same_length:
            raise ValueError(
                "The function to evaluate must produce an array with the same length as this dataset!"
            )
        return self.with_new_columns(**output)

    def filter(self, *masks: ColumnMask) -> Dataset:
        """
        Filter the dataset based on some criteria. See :ref:`Querying Based on Column
        Values` for more information.

        Parameters
        ----------
        *masks : Mask
            The masks to apply to dataset, constructed with :func:`opencosmo.col`

        Returns
        -------
        dataset : Dataset
            The new dataset with the masks applied.

        Raises
        ------
        ValueError
            If the given  refers to columns that are
            not in the dataset, or the  would return zero rows.

        """
        required_columns = set(m.column_name for m in masks)
        data = self.select(required_columns).get_data()
        bool_mask = np.ones(len(data), dtype=bool)
        for mask in masks:
            bool_mask &= mask.apply(data)

        new_state = self.__state.with_mask(bool_mask)
        return Dataset(self.__header, new_state, self.__tree)

    def rows(
        self,
        output="astropy",
        metadata_columns=[],
    ) -> Generator[Mapping[str, float | units.Quantity | np.ndarray]]:
        """
        Iterate over the rows in the dataset. Rows are returned as a dictionary
        For performance, it is recommended to first select the columns you need to
        work with.

        Parameters
        ----------
        output: str, default = "astropy"
            Whether to return values as "astropy" quantities or "numpy" scalars


        Yields
        -------
        row : dict
            A dictionary of values for each row in the dataset with units.

        """
        max = len(self)
        if max == 0:
            warn("Tried to iterate over a dataset with no rows!")

        chunk_ranges = [(i, min(i + 1000, max)) for i in range(0, max, 1000)]
        if len(chunk_ranges) == 0:
            raise StopIteration
        for start, end in chunk_ranges:
            chunk = self.take_range(start, end)
            chunk_data = chunk.get_data(output, metadata_columns=metadata_columns)
            try:
                output_chunk_data = dict(chunk_data)
            except TypeError:
                output_chunk_data = {self.columns[0]: chunk_data}

            if len(chunk) == 1:
                yield output_chunk_data
                return

            for i in range(len(chunk)):
                yield {k: v[i] for k, v in output_chunk_data.items()}

    def select(self, columns: str | Iterable[str]) -> Dataset:
        """
        Create a new dataset from a subset of columns in this dataset

        Parameters
        ----------
        columns : str or list[str]
            The column or columns to select.

        Returns
        -------
        dataset : Dataset
            The new dataset with only the selected columns.

        Raises
        ------
        ValueError
            If any of the given columns are not in the dataset.
        """
        new_state = self.__state.select(columns)
        return Dataset(
            self.__header,
            new_state,
            self.__tree,
        )

    def drop(self, columns: str | Iterable[str]) -> Dataset:
        """
        Create a new dataset without the provided columns.

        Parameters
        ----------
        columns : str or list[str]
            The columns to drop

        Returns
        -------
        dataset : Dataset
            The new dataset without the droppedcolumns

        Raises
        ------
        ValueError
            If any of the provided columns are not in the dataset.

        """
        if isinstance(columns, str):
            columns = [columns]

        current_columns = set(self.__state.columns)
        dropped_columns = set(columns)

        if missing := dropped_columns.difference(current_columns):
            raise ValueError(f"Columns {missing} are  not in this dataset")

        return self.select(current_columns - dropped_columns)

    def sort_by(self, column: str, invert: bool = False) -> Dataset:
        """
        Sort this dataset by the values in a given column. By default sorting is in
        ascending order (least to greatest). Pass invert = True to sort in descending
        order (greatest to least).

        This can be used to, for example, select largest halos in a given
        dataset:

        .. code-block:: python

            dataset = oc.open("haloproperties.hdf5")
            dataset = dataset
                        .sort_by("fof_halo_mass", invert=True)
                        .take(100, at="start")

        Parameters
        ----------
        column : str
            The column in the halo_properties or galaxy_properties dataset to
            order the collection by.

        invert : bool, default = False
            If False (the default), ordering will be from least to greatest.
            Otherwise greatest to least.

        Returns
        -------
        result : Dataset
            A new Dataset ordered by the given column.


        """
        new_state = self.__state.sort_by(column, invert)
        return Dataset(
            self.__header,
            new_state,
            self.__tree,
        )

    def take(
        self,
        n: int,
        at: str = "random",
    ) -> Dataset:
        """
        Create a new dataset from some number of rows from this dataset.

        Can take the first n rows, the last n rows, or n random rows
        depending on the value of 'at'.


        Parameters
        ----------
        n : int
            The number of rows to take.
        at : str
            Where to take the rows from. One of "start", "end", or "random".
            The default is "random".


        Returns
        -------
        dataset : Dataset
            The new dataset with only the selected rows.

        Raises
        ------
        ValueError
            If n is negative or greater than the number of rows in the dataset,
            or if 'at' is invalid.

        """

        new_state = self.__state.take(n, at)

        return Dataset(
            self.__header,
            new_state,
            self.__tree,
        )

    def take_range(self, start: int, end: int) -> Dataset:
        """
        Create a new dataset from a row range in this dataset. We use standard
        indexing conventions, so the rows included will be start -> end - 1.

        Parameters
        ----------
        start : int
            The beginning of the range
        end : int
            The end of the range

        Returns
        -------
        table : astropy.table.Table
            The table with only the rows from start to end.

        Raises
        ------
        ValueError
            If start or end are negative or greater than the length of the dataset
            or if end is greater than start.

        """
        new_state = self.__state.take_range(start, end)

        return Dataset(
            self.__header,
            new_state,
            self.__tree,
        )

    def take_rows(self, rows: np.ndarray | DataIndex):
        """
        Take the rows of a dataset specified by the :code:`rows` argument.
        :code:`rows` should be an array of integers.

        Parameters:
        -----------
        rows: np.ndarray[int]

        Returns
        -------
        dataset: The dataset with only the specified rows included

        Raises:
        -------
        ValueError:
            If any of the indices is less than 0 or greater than the length of the
            dataset.

        """
        if not isinstance(rows, np.ndarray):
            new_state = self.__state.take_rows(rows)
        else:
            index = SimpleIndex(rows)
            new_state = self.__state.take_rows(index)
        return Dataset(self.__header, new_state, self.__tree)

    def with_new_columns(
        self,
        descriptions: str | dict[str, str] = {},
        **new_columns: DerivedColumn | np.ndarray | units.Quantity,
    ):
        """
        Create a new dataset with additional columns. These new columns can be derived
        from columns already in the dataset, a numpy array, or an astropy quantity
        array. When a column is derived from other columns, it will behave
        appropriately under unit transformations. Columns provided directly as astropy
        quantities will not change under unit transformations. See
        :ref:`Adding Custom Columns` for examples.

        Parameters
        ----------

        descriptions : str | dict[str, str], optional
            A description for the new columns. These descriptions will be accessible through
            :py:attr:`Dataset.descriptions <opencosmo.Dataset.descriptions>`. If a dictionary,
            should have keys matching the column names.

        ** new_columns : opencosmo.DerivedColumn | np.ndarray | units.Quantity

        Returns
        -------
        dataset : opencosmo.Dataset
            This dataset with the columns added

        """
        if isinstance(descriptions, str):
            descriptions = {key: descriptions for key in new_columns.keys()}
        new_state = self.__state.with_new_columns(descriptions, **new_columns)
        return Dataset(self.__header, new_state, self.__tree)

    def make_schema(self, with_header: bool = True) -> DatasetSchema:
        """
        Prep to write the dataset. This should not be called directly for the user.
        The opencosmo.write file writer automatically handles the file context.

        Parameters
        ----------
        file : h5py.File
            The file to write to.
        dataset_name : str
            The name of the dataset in the file. The default is "data".

        """

        schema = self.__state.make_schema()
        if not with_header:
            schema.header = None

        if self.__tree is not None:
            tree = self.__tree.apply_index(self.__state.raw_index)
            spat_idx_schema = tree.make_schema()
            for name, column in spat_idx_schema.items():
                schema.add_child(column, name)
        return schema

    def with_units(
        self,
        convention: Optional[str] = None,
        conversions: dict[u.Unit, u.Unit] = {},
        **columns: u.Unit,
    ) -> Dataset:
        r"""
        Create a new dataset from this one with a different unit convention, and/or
        convert one unit to another across the entire dataset, or convert individual
        columns.

        Unit conversions are always performed after a change of convention, and
        changing conventions clears any existing unit conversions. Individual
        column conversions always take precedence over blanket unit conversions.

        Calling this function without arguments will clear any existing unit conversions.

        For more, see :doc:`units`.

        .. code-block:: python

            import astropy.units as u

            # this works
            dataset = dataset.with_units(fof_halo_mass=u.kg)

            # this clears the previous conversion
            dataset = dataset.with_units("scalefree")

            # This now fails, because the units of masses
            # are Msun / h, which cannot be converted to kg
            dataset = dataset.with_units(fof_halo_mass=u.kg)

            # this will work, the units of halo mass in the "physical"
            # convention are Msun (no h).
            dataset = dataset.with_units("physical", fof_halo_mass=u.kg, fof_halo_center_x=u.lyr)

            # Suppose you want all distances in lightyears, but the x coordinate of your
            # halo center in kilometers, for some reason ¯\_(ツ)_/¯
            blanket_conversions = {u.Mpc: u.lyr}
            dataset = dataset.with_units(conversions = blanket_conversions, fof_halo_center_x = u.km)



        Parameters
        ----------
        convention : str, optional
            The unit convention to use. One of "physical", "comoving",
            "scalefree", or "unitless".

        conversions : dict[astropy.units.Unit, astropy.Units.Unit]
            Conversions that apply to all columns in the dataset with the
            unit given by the key.

        **column_conversions: astropy.units.Unit
            Custom unit conversions for one or more or of the columns
            in this dataset.

        Returns
        -------
        dataset : Dataset
            The new dataset with the requested unit convention and/or conversions.

        """

        new_state = self.__state.with_units(
            convention, conversions, columns, self.cosmology, self.redshift
        )
        if convention is not None:
            new_header = self.__header.with_units(convention)
        else:
            new_header = self.__header

        return Dataset(
            new_header,
            new_state,
            self.__tree,
        )
